open Misc
open Error
open Parsetree
open Symbtree

exception Ctt_no_init of string * Location.t
exception Ctt_multiple_init of string * Location.t
exception Ctt_state_not_found of string * string * Location.t

let dual_direction = function
  | Dir_send -> Dir_receive
  | Dir_receive -> Dir_send

let ctt_find_init cid loc sl =
  let rec f i = function
  | [] ->
      (match i with
	 | Some i -> create_ident i
	 | None -> raise (Ctt_no_init (cid,loc)))
  | Cstate_init (s,_)::tl | Cstate_init_final (s,_)::tl ->
      (match i with
	 | None -> f (Some(s)) tl
	 | Some _ -> raise (Ctt_multiple_init(cid,loc)))
  | _::tl -> f i tl
  in f None sl

let ctt_find_finals sl =
  let rec f final = function
  | [] -> final
  | Cstate_final (s,_)::tl | Cstate_init_final (s,_)::tl ->
      f (IdSet.add (create_ident s) final) tl
  | _::tl -> f final tl
  in f IdSet.empty sl

let rec ctt_mk_send_transitions sl =
  match sl with
    | [] -> []
    | Cstate_init (s,tr)::tl | Cstate_final (s,tr)::tl
    | Cstate_init_final (s,tr)::tl | Cstate_base (s,tr)::tl ->
        let f = function
	  | Ctrans_send (mid,s') -> (create_ident s,mid,create_ident s')
	  | _ -> assert false
	and g = function Ctrans_send _ -> true | _ -> false
        in (List.rev_map f (List.filter g tr))@ctt_mk_send_transitions tl

let rec ctt_mk_receive_transitions sl =
  match sl with
    | [] -> []
    | Cstate_init (s,tr)::tl | Cstate_final (s,tr)::tl
    | Cstate_init_final (s,tr)::tl | Cstate_base (s,tr)::tl ->
        let f = function
	  | Ctrans_receive (mid,s') -> (create_ident s,mid,create_ident s')
	  | _ -> assert false
	and g = function Ctrans_receive _ -> true | _ -> false
        in (List.rev_map f (List.filter g tr))@ctt_mk_receive_transitions tl

let ctt_find ctts c =
  List.find (fun ci' -> ci'.cid=c) ctts

let ctt_walk ci d r a mid =
  let direction = if string_of_ident r = "server" then d else dual_direction d in
  let transitions = ci.transitions direction in
  let (_,_,b) = List.find (fun (a',m',_) -> a'=a && m'=mid) transitions in
    b
let ctt_walkall ci d r aset mid =
  let direction = if string_of_ident r = "server" then d else dual_direction d in
  let transitions = ci.transitions direction in
  let l = List.find_all (fun (a',m',_) -> (IdSet.mem a' aset && m'=mid)) transitions in
    List.fold_left (fun s -> fun (_,_,b) -> IdSet.add b s) IdSet.empty l

let ctt_is_final c st = 
  IdSet.mem st c.finals

type state_dir = Undef | Send | Receive

let ctt_check_positional c loc sl =
  let result = ref(true) in
  let error s =
    result := false;
    Error.print loc "WARNING: state %s of contract %s is not positional@." s c;
  in let f st = 
      let rec g s dir = function
	| [] -> ()
	| Ctrans_send _::tl ->
	    if dir = Receive then error s else g s Send tl
	| Ctrans_receive _::tl ->
	    if dir = Send then error s else g s Receive tl
      in match st with
	| Cstate_init (s, tl) | Cstate_final (s, tl)
	| Cstate_init_final (s, tl) | Cstate_base (s, tl) ->
	    g s Undef tl
  in List.iter f sl; !result

let ctt_check_deterministic c loc sl =
  let result = ref(true) in
  let error s =
    result := false;
    Error.print loc "WARNING: state %s of contract %s is not deterministic@." s c;
  in let f st = 
      let rec g s sl rl = function
	| [] -> ()
	| Ctrans_send (mid,_)::tl ->
	    if List.mem mid sl then error s else g s (mid::sl) rl tl
	| Ctrans_receive (mid,_)::tl ->
	    if List.mem mid rl then error s else g s sl (mid::rl) tl
      in match st with
	| Cstate_init (s, tl) | Cstate_final (s, tl)
	| Cstate_init_final (s, tl) | Cstate_base (s, tl) ->
	    g s [] [] tl
  in List.iter f sl; !result

let ctt_check_synchronizing c loc finals sl =
  let exists_cycle dir st  = 
    let followers st2 = 
      let filter trans = match trans with
	| Ctrans_send(_,st3) -> if (dir=Send) then [st3] else []
	| Ctrans_receive(_,st3) -> if (dir=Receive) then [st3] else []
      in let f = function
	| Cstate_init(id,_) | Cstate_final(id,_)
	| Cstate_init_final(id,_) | Cstate_base(id,_)
	    -> id=st2
      in let trans_list = match (List.find f sl) with
	| Cstate_init(_,tl) | Cstate_final(_,tl)
	| Cstate_init_final(_,tl) | Cstate_base(_,tl)
	    -> tl
      in List.flatten (List.map filter trans_list)
    in let rec reaches_st st2 marked = 
      if (st2=st) then true 
      else if List.mem st2 marked then false
      else let test st3 = reaches_st st3 (st2::marked)
      in List.exists test (followers st2) 
    in let initfollow=followers st
    in List.exists (fun st2->reaches_st st2 []) initfollow
  in not ((IdSet.exists (fun x -> exists_cycle Send (string_of_ident x)) finals)
	  or (IdSet.exists (fun x -> exists_cycle Receive (string_of_ident x)) finals))

let ctt_check_leakage c loc finals sl =
  let bp = ctt_check_positional c loc sl in
  let bd = ctt_check_deterministic c loc sl in
  let bs = ctt_check_synchronizing c loc finals sl in
    if not bp || not bd || not bs then
      Format.fprintf !Config.formatter
	"@.WARNING: program cannot be guaranteed leak-free because of contract %s@." c

let ctrans_to_dot_stmt s = function
  | Ctrans_send (m,s') ->
      s ^ " -> " ^ s' ^ "[label=\"!" ^ m ^ "\"];\n"
  | Ctrans_receive (m,s') ->
      s ^ " -> " ^ s' ^ "[label=\"?" ^ m ^ "\"];\n"

let cstate_to_dot_stmt = function
  | Cstate_init (s, tl) ->
      (s ^ " [shape=box];\n")::(List.map (ctrans_to_dot_stmt s) tl)
  | Cstate_final (s, tl) ->
      (s ^ " [shape=doublecircle];\n")::(List.map (ctrans_to_dot_stmt s) tl)
  | Cstate_init_final (s, tl) ->
      (s ^ " [shape=doubleoctagon];\n")::(List.map (ctrans_to_dot_stmt s) tl)
  | Cstate_base (s, tl) ->
      (s ^ ";\n")::(List.map (ctrans_to_dot_stmt s) tl)

let ctt_view_dot c sl =
  let attr_graph = "graph [rankdir=LR];" in
  let attr_node = "node [shape=ellipse, color=gray];" in
  let attr_edge = "edge [];" in
  let attr = attr_graph ^ "\n" ^ attr_node ^ "\n" ^ attr_edge ^ "\n" in
  let g =
    (List.fold_left (^)
       ("Digraph "^c^" {\n"^ attr)
       (List.flatten (List.map cstate_to_dot_stmt sl))
    )^"}\n" in
  let file = open_out ("output-"^c^".dot") in
    output_string file g; close_out file;
    if !Config.view_contracts then (
      let output = "output-"^c^".dot" in
      let dot_output = "output-"^c^"."^(!Config.dot_type) in
      let dot_cmd = "dot -T"^(!Config.dot_type)^" "^output^" > "^dot_output in
      let viewer_cmd = (!Config.viewer)^" "^dot_output^" >& /dev/null &" in
	Format.fprintf !Config.formatter "@.Contract %s in %s@." c output;
	let _ = Sys.command (dot_cmd^" && "^viewer_cmd)
      in ())
